__doc__ == """
**naclo** is a Python cleaning toolset for small molecule drug discovery datasets.
"""

from naclo.mol_stats import *
from naclo.mol_conversion import *
from naclo import database
from naclo import dataframes
from naclo import fragments
from naclo import neutralize
from naclo import rdpickle
from naclo.Bleach import Bleach
from naclo.Binarize import Binarize
from naclo.__asset_loader import bleach_default_model
from naclo.__asset_loader import binarize_default_params, binarize_default_options
from naclo.UnitConverter import UnitConverter
from naclo import __naclo_util
from naclo import __model_schema



from schema import Schema, Or


bleach_schema = Schema({
    # Structure data
    'structure_col': str,
    'structure_type': Or('smiles', 'mol'),

    # Target data
    'target_col': Or(str, None),
    'input_units': Or('mixed', '-log(M)', 'M', 'mM', 'µM', 'uM', 'nM', ''),
    'output_units': Or('-log(M)', 'M', 'mM', 'µM', 'uM', 'nM', ''),
    'units_col': str,
    'drop_na_units': bool,

    # Formatting
    'format_settings': {
        'duplicate_compounds': Or('average', 'remove', 'keep'),
        'append_columns': [Or('smiles', 'mol', 'inchi_key', 'mw')],
        'remove_header_chars': str,
        'remove_na_targets': bool
    },

    # Chemistry
    'chem_settings': {
        'neutralize_charge': bool,
        'desalt': bool,
        'fragment_filter': Or('carbon_count', 'atom_count', 'mw', 'none', '')
    }
})



import warnings

import stse


def drop_na_structures(self) -> None:
        """Drops NA along declared structure column."""
        self.df.dropna(subset=[self.structure_col], inplace=True)
        if not len(self.df):
            warnings.warn('ALL_NA_STRUCTURES: All structures in specified column were NA, all rows dropped',
                          RuntimeWarning)

def drop_na_targets(self) -> None:
    """Drops NA along declared target column"""
    run_na_targets = self.file_settings['remove_na_targets']['run']

    if self.target_col and run_na_targets and len(self.df):  # If run and TARGET COLUMN DECLARED
        self.df.dropna(subset=[self.target_col], inplace=True)
        if not len(self.df):
            warnings.warn('ALL_NA_TARGETS: All targets in specified column were NA, all rows dropped',
                            RuntimeWarning)

    elif run_na_targets:  # If run but not declared target
        warnings.warn('NA_TARGETS: options.file_settings.remove_na_targets was set to run but no activity column \
            was specified', RuntimeWarning)
        
def recognized_options_checker(input_options, recognized_options) -> None:  # *
    input = stse.dictionaries.branches(input_options)
    recognized = stse.dictionaries.branches(recognized_options)

    errors = {}
    for key, value in recognized.items():
        if isinstance(value, list):
            if not input[key] in recognized[key]:
                errors[f'BAD_OPTION{key.upper()}'] = f'"{input[key]}" is not an accepted value for "{key}", set \
                    to one of: "{recognized[key]}"'
        else:
            if not type(input[key]) == type(recognized[key]):
                errors[f'BAD_OPTION{key.upper()}'] = f'{type(input[key])} is not an accepted type for {key}, \
                    input a {type(recognized[key])}'
    if errors:
        raise ValueError(errors)
    
def check_columns_in_df(df, columns) -> None:
    """Checks that all columns in columns are in df"""
    for column in columns:
        if column not in df.columns:
            raise ValueError(f'Column: "{column}" is not found in data.')



import unittest
import numpy as np
import pandas as pd

import naclo
from naclo import UnitConverter


class TestUnitConverter(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.test_values = [
            55,
            4,
            7,
            100,
            2000
        ]
        cls.test_units = [
            'ug•ml-1',
            'mg/l',
            'unrecognized',
            np.nan,
            'pm'
        ]
        cls.test_mws = [
            220,
            400,
            300,
            110,
            150
        ]
        cls.test_smiles = 5*['CCC']
        cls.test_mols = naclo.smiles_2_mols(cls.test_smiles)
        
        cls.unit_converter = UnitConverter(cls.test_values, cls.test_units, cls.test_mws)
        return super().setUpClass()
    
    def test_to_molar_conversions(self):
        expected = np.array([
            250e-6,
            10e-6,
            np.nan,
            np.nan,
            2000e-12
        ])
        
        # Test all molar prefixes
        molars = self.unit_converter.to_molar()
        milli_molars = self.unit_converter.to_milli_molar()
        micro_molars = self.unit_converter.to_micro_molar()
        nano_molars = self.unit_converter.to_nano_molar()
        for vals, multi in zip([molars, milli_molars, micro_molars, nano_molars],
                               [1, 1e3, 1e6, 1e9]):
            self.assertIsInstance(
                vals,
                pd.Series
            )
            self.assertTrue(
                np.allclose(
                    vals.to_numpy(),
                    expected * multi,
                    equal_nan=True  # np.nan does not evaluate equal
                )
            )
        
    def test_to_neg_log_molar(self):
        neg_log_molars = self.unit_converter.to_neg_log_molar()
        
        self.assertIsInstance(
            neg_log_molars,
            pd.Series
        )
        
        expected = np.array([
            3.60206,
            5,
            np.nan,
            np.nan,
            8.69897
        ])
        
        self.assertTrue(
            np.allclose(
                neg_log_molars.to_numpy(),
                expected,
                equal_nan=True
            )
        )
        
    def test_df_converter(self):
        # Dont drop NA
        df = pd.DataFrame({'value': self.test_values, 'unit': self.test_units, 'mol': self.test_mols})
        out = UnitConverter.df_converter(df, 'mol', 'value', 'unit', 'm', drop_na_units=False)
        
        self.assertIn(
            'molar_value',
            out.columns
        )
        self.assertEqual(
            out['molar_value'].iloc[-1],
            df['value'].iloc[-1] * 1e-12  # pM
        )
        
        # Drop NA indices
        out = UnitConverter.df_converter(df, 'mol', 'value', 'unit', 'm', drop_na_units=True)
        self.assertEqual(
            list(out.index),
            [0, 1, 4]
        )
        
    def test_convert(self):
        # Test molar values --> -log(M) w/ case insensitivity
        units = ['M', 'mM', 'uM', 'nM', 'µM']
        for unit in units + [u.lower() for u in units] + [u.upper() for u in units[:-1]]:  # DONT CAP THE MU
            out = UnitConverter.convert(self.test_values, unit, '-log(m)')
            self.assertIsInstance(out, list)
            
        # Test g/L values
        g_ovr_l_units = ['g/l', 'mg/L', 'ug/L', 'ng/L']
        g_ovr_l_test_values = [5]
        test_mws = [100]
        molars = [5/100, 5e-3/100, 5e-6/100, 5e-9/100]
        
        for u, m in zip(g_ovr_l_units, molars):
            self.assertAlmostEqual(
                UnitConverter.convert(g_ovr_l_test_values, u, 'm', test_mws)[0],
                m
            )
        
        # No change
        self.assertEqual(
            UnitConverter.convert(g_ovr_l_test_values, 'g/l', 'g/l', test_mws),
            g_ovr_l_test_values
        )
        
        # print(self.test_values)
        # mws = []
        # for unit in units + [u.lower() for u in units] + [u.upper() for u in units[:-1]]:
        
        # ERROR: invalid input unit
        with self.assertRaises(ValueError):
            UnitConverter.convert(self.test_values, 'unrecognized', '-log(m)')
        
        # ERROR: invalid output unit
        with self.assertRaises(ValueError):
            UnitConverter.convert(self.test_values, 'm', 'unrecognized')

        
if __name__ == '__main__':
    unittest.main()



from importlib import resources
import json
import yaml
from yaml.loader import SafeLoader


with resources.open_text('naclo.assets', 'bleach_model.yml') as f:
    bleach_default_model = yaml.load(f, Loader=SafeLoader)

with resources.open_text('naclo.assets', 'binarize_default_params.json') as f:
    binarize_default_params = json.load(f)

with resources.open_text('naclo.assets', 'binarize_default_options.json') as f:
    binarize_default_options = json.load(f)

with resources.open_text('naclo.assets', 'recognized_bleach_options.json') as f:
    recognized_bleach_options = json.load(f)
    
with resources.open_text('naclo.assets', 'recognized_binarize_options.json') as f:
    recognized_binarize_options = json.load(f)

with resources.open_text('naclo.assets', 'recognized_units.json') as f:
    recognized_units = json.load(f)

with resources.open_text('naclo.assets', 'recognized_salts.json') as f:
    recognized_salts = json.load(f)
    
with resources.open_text('naclo.assets', 'bleach_warnings.yml') as f:
    bleach_warnings = yaml.load(f, Loader=SafeLoader)



from logging import warning
from math import log10
import pandas as pd
import numpy as np
import copy
from typing import Union, Iterable

import naclo
from naclo.__asset_loader import recognized_units


class UnitConverter:
    _recognized_units = copy.deepcopy(recognized_units)
    _standard_unit_col = 'standard_unit'
    _multiplier_col = 'multiplier'
    
    # Unit groups
    _molars = [
        'pm',
        'nm',
        'um',
        'µm',  # Both forms of micro work
        'mm',
        'm'
    ]
    
    _g_ovr_ls = [
        'pg/l',
        'ng/l',
        'ug/l',
        'mg/l',
        'g/l',
        'pg/ml',
        'ng/ml',
        'ug/ml',
        'mg/ml',
        'g/ml'
    ]
    
    def __init__(self, values:Iterable, units:Iterable, mol_weights:Iterable) -> None:
        # Column names
        self._unit_col = 'unit'
        self._value_col = 'value'
        self._mw_col = 'mol_weight'
        
        self.df = pd.DataFrame({
            self._unit_col: copy.deepcopy(units),
            self._value_col: copy.deepcopy(values),
            self._mw_col: copy.deepcopy(mol_weights)
        })
        
        self.df = UnitConverter.standardize_units(self.df, self._unit_col)
    
    @staticmethod
    def neg_log(val:Union[int, float, np.number]) -> float:
        """Simple negative log10 computation.

        Args:
            val (Union[int, float, np.number]): Value to use for computation.

        Returns:
            float: Computed value.
        """
        if val < 0:
            warning(f'Negative value {val} passed to UnitConverter.__neg_log().')
            return np.nan
        elif val == 0:
            return 0
        else:
            return -1*log10(val)
    
    @staticmethod
    def inv_neg_log(val:float) -> float:
        '''Inverse of self.neg_log()'''
        return 10**(-1*val)
        
    @staticmethod
    def df_converter(df:pd.DataFrame, mol_col_name:str, value_col_name:str, units_col_name:str, output_units:str,
                         drop_na_units:bool) -> pd.DataFrame:
        """Converts mixed units to desired units using unit column in df

        Args:
            df (pd.DataFrame): Dataset
            mol_col_name (str): Name of Mol column in df
            value_col_name (str): Name of numerical value column in df refered to by units
            units_col_name (str): Name of units column in df
            output_units (str): Output unit type
            drop_na_units (bool): If True, drops w/ NA units. Else, keeps the values as nan

        Returns:
            pd.DataFrame: Dataset w/ units converted
        """
        df = df.copy()
        uc = UnitConverter(df[value_col_name], df[units_col_name], naclo.mol_weights(df[mol_col_name]))
        output_units = output_units.lower()
        
        if output_units == '-log(m)':
            col_name = f'neg_log_molar_{value_col_name}'
            df[col_name] = uc.to_neg_log_molar()
        elif output_units == 'm':
            col_name = f'molar_{value_col_name}'
            df[col_name] = uc.to_molar()
        elif output_units == 'mm':
            col_name = f'milli_molar_{value_col_name}'
            df[col_name] = uc.to_milli_molar()
        elif output_units in ['um', 'µm']:
            col_name = f'micro_molar_{value_col_name}'
            df[col_name] = uc.to_micro_molar()
        elif output_units == 'nm':
            col_name = f'nano_molar_{value_col_name}'
            df[col_name] = uc.to_nano_molar()
        else:
            raise ValueError(f'Output units: "{output_units}" are not recognized')
        
        return df.dropna(subset=[col_name]) if drop_na_units else df
    
    @staticmethod
    def convert(values:Iterable[float], in_unit:str, out_unit:str, mws:Iterable[float]=None) -> str:
        '''Simply converts from one unit to another. out_unit can be -log(M)'''
        if in_unit.lower() == out_unit.lower():
            return values
        
        for u in [in_unit, out_unit]:
            if u.lower() not in ['-log(m)'] + UnitConverter._molars + UnitConverter._g_ovr_ls:
                raise ValueError(f'Unit {u} not recognized')
            elif not mws and u.lower() in UnitConverter._g_ovr_ls:
                raise ValueError(f'Unit {u} requires molar weights')
        
        if in_unit.lower() == '-log(m)':
            molar_values = [UnitConverter.inv_neg_log(v) for v in values]
        elif in_unit.lower() in UnitConverter._molars:
            molar_values = [v * UnitConverter._recognized_units[in_unit.lower()][1] for v in values]
        else:
            molar_values = [v / mw * UnitConverter._recognized_units[in_unit.lower()][1] for v, mw in zip(values, mws)]
        
        if out_unit.lower() == '-log(m)':
            return [UnitConverter.neg_log(m) for m in molar_values]
        elif out_unit.lower() in UnitConverter._molars:
            return [m / UnitConverter._recognized_units[out_unit.lower()][1] for m in molar_values]
        else:
            return [(m*mw) / UnitConverter._recognized_units[out_unit.lower()][1] for m, mw in zip(molar_values, mws)]
    
    @staticmethod
    def standardize_units(df, unit_col_name) -> pd.DataFrame:
        """Appends standard units and multipliers found in naclo/assets/recognized_units.json to self.df. Appends
        np.nan if unit is not recognized."""
        standard_units = []
        multipliers = []
        
        for unit in df[unit_col_name]:
            try:
                standard_unit, multiplier = UnitConverter._recognized_units[f'{unit}'.lower()]
                standard_units.append(standard_unit)
                multipliers.append(multiplier)
            except KeyError:
                standard_units.append(np.nan)
                multipliers.append(np.nan)
                continue
            
        df[UnitConverter._standard_unit_col] = standard_units
        df[UnitConverter._multiplier_col] = multipliers
        return df
            
    def __to_molar_broadcaster(self, row:pd.Series) -> float:
        """Broadcasting function to use with pd.DataFrame.apply(). Computes molar values from self.df.

        Args:
            row (pd.Series): pd.DataFrame row passed in apply().

        Returns:
            Union[float]: Molar value. np.nan if the standard unit is not found in self.molar or self.g_ovr_l.
        """
        if row[self._standard_unit_col] in UnitConverter._molars:
            molar_val = float(row[self._value_col])*row[self._multiplier_col]
        elif row[self._standard_unit_col] in UnitConverter._g_ovr_ls:  # Divide by mw (g/mol) -> g/L * mol/g = M
            molar_val = float(row[self._value_col])*row[self._multiplier_col]/row[self._mw_col]  # Divide by mw
        else:
            molar_val = np.nan
        return molar_val
            
    def to_neg_log_molar(self) -> pd.Series:
        """Applies self.__to_molar_broadcaster and self.__neg_log to self.df.

        Returns:
            pd.Series: Negative log molar values.
        """
        return self.to_molar().apply(self.neg_log)
    
    def to_molar(self) -> pd.Series:
        """Applies self.__to_molar_broadcaster to self.df.

        Returns:
            pd.Series: Molar values.
        """
        return self.df.apply(self.__to_molar_broadcaster, axis=1)
    
    def to_milli_molar(self) -> pd.Series:
        return self.to_molar().apply(lambda x: x*1e3)

    def to_micro_molar(self) -> pd.Series:
        return self.to_molar().apply(lambda x: x*1e6)
    
    def to_nano_molar(self) -> pd.Series:
        return self.to_molar().apply(lambda x: x*1e9)



import unittest
import copy
import numpy as np
import pandas as pd
from rdkit import Chem
from math import log10

import naclo
from naclo import Bleach
from naclo import UnitConverter
from naclo.__asset_loader import bleach_default_model
import stse


class TestBleach(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        cls._smiles_df = pd.DataFrame({
            'SMILES': ['Cc1cc(/C=C/C#N)cc(C)c1Nc1nc(Nc2ccc(C#N)cc2)ncc1N',
                       'Cc1cc(/C=C/C#N)cc(C)c1Nc1ncc(N)c(Nc2c(C)cc(/C=C/C#N)cc2C)n1',
                       'CCC',
                       'CCC.Cl',
                       'C.NO.S',
                       'Br',
                       '']
        })
        return super().setUpClass()
    
    @property
    def smiles_df(self):
        return self._smiles_df.copy()
    
    @property
    def default_model(self):
        return copy.deepcopy(bleach_default_model)
    
    def test_drop_na(self):
        df = pd.DataFrame({
            'SMILES': [
                pd.NA,
                '',
                'nan',
                'none',
                'CCC',
                np.nan
            ],
            'ROMol': [
                pd.NA,
                '',
                'nan',
                'C',
                None,
                np.nan
            ],
            'drop_empty': 6*[np.nan]
        })
        model = self.default_model
        
        # SMILES
        model['structure_col'] = 'SMILES'
        model['structure_type'] = 'smiles'
        
        bleach = Bleach(df, model)
        bleach.drop_na()
        
        expected = pd.DataFrame({
            'SMILES': ['CCC']
        }, index=[4])
        
        self.assertTrue(bleach.df.equals(expected))
        
        # Mols
        model['structure_col'] = 'ROMol'
        model['structure_type'] = 'mol'
        
        bleach = Bleach(df, model)
        bleach.drop_na()
        
        expected = pd.DataFrame({
            'ROMol': ['C']
        }, index=[3])
        
        self.assertTrue(bleach.df.equals(expected))
        
        # ALL_NA_STRUCTURES warning
        model['structure_col'] = 'drop_empty'
        bleach = Bleach(df, model)
        
        with self.assertWarns(RuntimeWarning):
            bleach.drop_na()
        self.assertTrue(bleach.df.equals(pd.DataFrame()))  # Empty df, all rows dropped
        
        # ALL_NA_TARGETS warning
        model['structure_col'] = 'ROMol'
        model['target_col'] = 'drop_empty'
        
        # Not set to drop NA targets
        bleach = Bleach(df, model)
        bleach.drop_na()
        
        expected = pd.DataFrame({
            'ROMol': ['C']
        }, index=[3])
        
        self.assertTrue(bleach._df.equals(expected))
        
        # Set to drop NA targets
        model['format_settings']['remove_na_targets'] = True
        bleach = Bleach(df, model)
        
        with self.assertWarns(RuntimeWarning):
            bleach.drop_na()
        self.assertTrue(bleach.df.equals(pd.DataFrame()))  # Empty df, all rows dropped
        
    def test_init_structure_compute(self):
        # From SMILES
        model = self.default_model
        model['structure_col'] = 'SMILES'
        model['structure_type'] = 'smiles'
        
        bleach = Bleach(self.smiles_df, model)
        bleach.drop_na()
        bleach.init_structure_compute()

        # Mol instances
        [self.assertIsInstance(m, Chem.rdchem.Mol) for m in bleach._df['ROMol']]
        
        # From SMILES w/ bad Mols
        mol_df = bleach.df
        
        new_row = pd.DataFrame({
            'SMILES': ['CC', 'C', 'CCC'],
            'ROMol': ['test', 1, 'CCC']
        })
        
        mol_df = pd.concat((mol_df, new_row))
        bleach = Bleach(mol_df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        
        # Rebuilt SMILES are the same
        self.assertEqual(
            bleach.df['SMILES'].tolist(),
            mol_df['SMILES'].tolist()
        )
        [self.assertIsInstance(m, Chem.rdchem.Mol) for m in bleach._df['ROMol']]
        
        # From Mols w/ bad Mols
        model['structure_col'] = 'ROMol'
        model['structure_type'] = 'mol'
        bleach = Bleach(mol_df, model)

        bleach.drop_na()
        bleach.init_structure_compute()
        
        # Drop bad Mols
        [self.assertIsInstance(m, Chem.rdchem.Mol) for m in bleach._df['ROMol']]
        self.assertEqual(
            len(bleach.df),
            len(mol_df) - 3
        )
        self.assertEqual(
            bleach._df['SMILES'].tolist(),
            self.smiles_df['SMILES'].tolist()[:-1]  # Excluding blank
        )
        
    def test_chem_cleanup(self):
        model = self.default_model
        model['structure_col'] = 'SMILES'
        model['structure_type'] = 'smiles'
        
        # Desalt only
        model['chem_settings']['desalt'] = True
        model['chem_settings']['fragment_filter'] = 'none'
        model['chem_settings']['neutralize_charge'] = False
        
        bleach = Bleach(self.smiles_df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        
        expected_smiles = [
            'Cc1cc(/C=C/C#N)cc(C)c1Nc1nc(Nc2ccc(C#N)cc2)ncc1N',
            'Cc1cc(/C=C/C#N)cc(C)c1Nc1ncc(N)c(Nc2c(C)cc(/C=C/C#N)cc2C)n1',
            'CCC',
            'CCC',
            'C.NO.S'
        ]
        
        self.assertEqual(
            bleach.df['SMILES'].tolist(),
            expected_smiles
        )
        
        # Filter only
        model['chem_settings']['desalt'] = False
        
        ## carbon count
        model['chem_settings']['fragment_filter'] = 'carbon_count'
        
        bleach = Bleach(self.smiles_df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        
        self.assertEqual(
            bleach.df['SMILES'].iloc[-2],
            'C'
        )
        
        ## atom count
        model['chem_settings']['fragment_filter'] = 'atom_count'
        
        bleach = Bleach(self.smiles_df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        
        self.assertEqual(
            bleach.df['SMILES'].iloc[-2],
            'NO'
        )
        
        ## molecular weight
        model['chem_settings']['fragment_filter'] = 'mw'
        
        bleach = Bleach(self.smiles_df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        
        self.assertEqual(
            bleach.df['SMILES'].iloc[-2],
            'S'
        )
        
        # Salts + filter together
        model['chem_settings']['desalt'] = True
        model['chem_settings']['fragment_filter'] = 'carbon_count'
        model['chem_settings']['neutralize_charge'] = False
        
        bleach = Bleach(self.smiles_df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        
        expected_smiles = [
            'Cc1cc(/C=C/C#N)cc(C)c1Nc1nc(Nc2ccc(C#N)cc2)ncc1N',
            'Cc1cc(/C=C/C#N)cc(C)c1Nc1ncc(N)c(Nc2c(C)cc(/C=C/C#N)cc2C)n1',
            'CCC',
            'CCC',
            'C'
        ]
        
        self.assertEqual(
            bleach.df['SMILES'].tolist(),
            expected_smiles
        )
        
    def test_handle_duplicates(self):
        df = self.smiles_df
        df['target'] = [1, 2, 3, 4, 5, 6, 7]
        
        model = self.default_model
        model['structure_col'] = 'SMILES'
        model['structure_type'] = 'smiles'
        model['target_col'] = 'target'
        
        expected = pd.DataFrame({
            'SMILES': ['Cc1cc(/C=C/C#N)cc(C)c1Nc1nc(Nc2ccc(C#N)cc2)ncc1N',
                       'Cc1cc(/C=C/C#N)cc(C)c1Nc1ncc(N)c(Nc2c(C)cc(/C=C/C#N)cc2C)n1',
                       'CCC',
                       'C'],
            
            'InchiKey': [
                'NFQWIYBXPYHNRC-ONEGZZNKSA-N',
                'IKTVSGAURLVDFP-KQQUZDAGSA-N',
                'ATUOYWHBWRKTHZ-UHFFFAOYSA-N',
                'VNWKTOKETHGBQD-UHFFFAOYSA-N'],
            
            'target': [1,
                       2,
                       3.5,
                       5]
        })
        
        # Remove
        model['format_settings']['duplicate_compounds'] = 'remove'
        
        bleach = Bleach(df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        bleach.handle_duplicates()
        
        expected_remove = expected.copy()
        expected_remove['target'] = [1, 2, 3, 5]
        
        self.assertTrue(bleach.df[['SMILES', 'InchiKey', 'target']].reset_index(drop=True).equals(expected_remove))
        
        # Keep
        model['format_settings']['duplicate_compounds'] = 'keep'
        
        bleach = Bleach(df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        
        df_before = bleach.df
        bleach.handle_duplicates()
        df_after = bleach.df.drop('InchiKey', axis=1)
        
        self.assertTrue(df_before.equals(df_after))
        
        self.assertIn(  # Added InChI key column
            'InchiKey',
            bleach.df.columns
        )
        
        # Average
        model['format_settings']['duplicate_compounds'] = 'average'
        
        ## WARNING: no target column
        model['target_col'] = ''
        
        bleach = Bleach(df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        
        with self.assertWarns(UserWarning):
            bleach.handle_duplicates()
        ### defaults to same as remove
        self.assertTrue(bleach.df[['SMILES', 'InchiKey', 'target']].reset_index(drop=True).equals(expected_remove))
        
        ## ERROR: input units == mixed but units column not specified
        model['target_col'] = 'target'
        model['input_units'] = 'mixed'
        
        bleach = Bleach(df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        
        with self.assertRaises(ValueError):
            bleach.handle_duplicates()
            
        ## WARNING: input units != mixed but units column specified (even though not even present in df)
        model['target_col'] = 'target'
        model['input_units'] = 'uM'
        model['units_col'] = 'units'
        
        expected = naclo.dataframes.df_smiles_2_inchi_keys(df, 'SMILES', 'InchiKey')
        expected['target'] = [-1 * log10(i * 1e-6) for i in df['target']]
        expected = stse.duplicates.average(expected, subsets=['InchiKey'], average_by='target')
        expected_avg_targets = expected['target'].tolist()[:2] + [expected['target'].iloc[2:4].mean()] + \
            [expected['target'].tolist()[4]]
        
        bleach = Bleach(df, model)
        
        ### check pointer before (should be unchanged)
        self.assertEqual(
            bleach._target_col,
            model['target_col'],
            'target'
        )
        self.assertEqual(
            bleach._target_units,
            model['input_units'],
            'uM'
        )
        
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        
        with self.assertWarns(UserWarning):
            bleach.handle_duplicates()
            
        ### check pointer after (should be changed to -log(M))
        self.assertEqual(
            bleach._target_col,
            'neg_log_molar_' + model['target_col'],
            'neg_log_molar_target'
        )
        self.assertEqual(
            bleach._target_units,
            '-log(m)'
        )
            
        model['units_col'] = ''
        bleach_2 = Bleach(df, model)
        bleach_2.drop_na()
        bleach_2.init_structure_compute()
        bleach_2.chem_cleanup()
        bleach_2.handle_duplicates()
            
        self.assertEqual(
            bleach.df['target'].tolist(),  # Static converter (µM) in response to warning
            bleach_2.df['target'].tolist(),  # No units column at all
            expected_avg_targets
        )
        
        ## mixed units
        model['input_units'] = 'mixed'
        model['units_col'] = 'units'
        df['units'] = [np.nan, 'nm', 'm', 'pg ml-1', 'UM', 'mM', 'M']
        
        bleach = Bleach(df, model)
        
        ### check pointer before (should be unchanged)
        self.assertEqual(
            bleach._target_col,
            model['target_col'],
            'target'
        )
        self.assertEqual(
            bleach._target_units,
            model['input_units'],
            'mixed'
        )
        
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        bleach.handle_duplicates()
        
        ### check pointer after (should be changed to -log(M))
        self.assertEqual(
            bleach._target_col,
            'neg_log_molar_' + model['target_col'],
            'neg_log_molar_target'
        )
        self.assertEqual(
            bleach._target_units,
            '-log(m)'
        )
        
    def test_final_unit_conversion(self):
        df = self.smiles_df
        df['target'] = [1, 2, 3, 4, 5, 6, 7]
        
        model = self.default_model
        model['structure_col'] = 'SMILES'
        model['structure_type'] = 'smiles'
        model['target_col'] = 'target'
        
        model['input_units'] = 'mixed'
        model['output_units'] = '-log(M)'
        model['units_col'] = 'units'
        df['units'] = [np.nan, 'nm', 'm', 'pg ml-1', 'UM', 'mM', 'M']
        
        bleach = Bleach(df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        bleach.handle_duplicates()
        
        # No change: AVERAGE therefore -log(M) -> -log(M)
        before = bleach.df
        bleach.final_unit_conversion()
        after = bleach.df
        
        self.assertTrue(before.equals(after))
        
        model['output_units'] = 'uM'
        bleach = Bleach(df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        bleach.handle_duplicates()
        
        before = bleach.df
        bleach.final_unit_conversion()
        after = bleach.df
        
        self.assertFalse(before.equals(after))
        
        ## correct conversion factor
        u_molar = [1e6 * UnitConverter.inv_neg_log(v) for v in before['neg_log_molar_target']]
        self.assertTrue(
            np.allclose(
                u_molar,
                after['{0}_{1}'.format(model['output_units'], model['target_col'])].tolist(),
                equal_nan=True
            )
        )
        
        ## retains all conversion columns
        self.assertIn('target', after.columns)
        self.assertIn('neg_log_molar_target', after.columns)
        self.assertIn('uM_target', after.columns)
        
        # Not average: therefore no initial -log(M) conversion
        df = self.smiles_df
        df['target'] = [1, 2, 3, 4, 5, 6, 7]
        
        model = self.default_model
        model['structure_col'] = 'SMILES'
        model['structure_type'] = 'smiles'
        model['target_col'] = 'target'
        
        model['format_settings']['duplicate_compounds'] = 'remove'
        model['input_units'] = 'M'  # NOTE: For easy conversion factor testing
        model['output_units'] = '-log(M)'
        model['units_col'] = 'units'
        
        ## -> -log(M) conversion
        bleach = Bleach(df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        bleach.handle_duplicates()
        
        before = bleach.df
        
        self.assertNotIn('neg_log_molar_target', before.columns)  # No -log(M)
        self.assertIn('target', before.columns)
        
        with self.assertWarns(RuntimeWarning):  # Units column specified but input units not mixed
            bleach.final_unit_conversion()
        after = bleach.df
        
        self.assertIn('neg_log_molar_target', after.columns)  # -log(M)
        self.assertIn('target', after.columns)
        
        ### correct conversion factor
        neg_log_molars = np.array([UnitConverter.neg_log(v) for v in before['target']])  # NOTE: Easy conversion factor!
        self.assertEqual(
            neg_log_molars.round(decimals=4).tolist(),
            after['neg_log_molar_target'].to_numpy().round(decimals=4).tolist()
        )
        
    def test_append_columns(self):
        model = self.default_model
        model['structure_col'] = 'SMILES'
        model['structure_type'] = 'smiles'
        model['input_units'] = 'M'
        model['output_units'] = 'M'  # Same units, not conversion to avoid no target col warning
        model['format_settings']['duplicate_compounds'] = 'remove'  # Avoid no target col warning
        
        # All columns
        bleach = Bleach(self.smiles_df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        bleach.handle_duplicates()
        bleach.final_unit_conversion()
        bleach.append_columns()
        
        self.assertEqual(
            ['SMILES', 'ROMol', 'InchiKey', 'MW'],  # NOTE: No taget so dont worry about
            list(bleach._df.columns)
        )
        
        # No columns (except SMILES for testing purposes)
        model['format_settings']['append_columns'] = ['smiles']
        
        bleach = Bleach(self.smiles_df, model)
        bleach.drop_na()
        bleach.init_structure_compute()
        bleach.chem_cleanup()
        bleach.handle_duplicates()
        bleach.final_unit_conversion()
        bleach.append_columns()
        
        self.assertEqual(
            ['SMILES'],
            list(bleach._df.columns)
        )
        
    def test_remove_header_chars(self):
        model = self.default_model
        model['structure_col'] = 'SMILES'
        model['structure_type'] = 'smiles'
        model['input_units'] = 'M'
        model['output_units'] = 'M'  # Same units, not conversion to avoid no target col warning
        model['format_settings']['duplicate_compounds'] = 'remove'  # Avoid no target col warning
        
        model['format_settings']['remove_header_chars'] = 'sr'
        
        bleach = Bleach(self.smiles_df, model)
        bleach.main()
        
        self.assertEqual(
            ['MILE', 'OMol', 'InchiKey', 'MW'],
            list(bleach._df.columns)
        )
        
    
if __name__ == '__main__':
    unittest.main()



import pandas as pd
import warnings
from typing import Callable, Iterable, Optional
import copy

# sourced from github.com/jwgerlach00
import naclo
import stse
from naclo.UnitConverter import UnitConverter
from naclo.__asset_loader import bleach_default_model, bleach_warnings
from naclo.__model_schema import bleach_schema

class Bleach:
    structure_types = sorted(['smiles', 'mol'])
    
    input_units = [
        'mixed',
        '-log(M)',
        'M',
        'mM',
        'µM',
        'nM',
        'g/L',
        'mg/L'
    ]
    
    output_units = [  # No mixed
        '-log(M)',
        'M',
        'mM',
        'µM',
        'nM',
        'g/L',
        'mg/L'
    ]
    
    duplicate_methods = sorted(['average', 'remove', 'keep'])
    
    fragment_filters = sorted(['carbon_count', 'mw', 'atom_count', 'none'])
    
    def __init__(self, df:pd.DataFrame, model:dict=bleach_default_model) -> None:  # *
        # Load user options
        self._model = copy.deepcopy(model)
        
        # Validate model
        bleach_schema.validate(self.model)

        self._default_cols = {
            'smiles': 'SMILES',
            'mol': 'ROMol',
            'inchi_key': 'InchiKey',
            'mw': 'MW'
        }

        # Save user input data
        self._original_df = df.copy()
        self._df = df.copy()

        self.mol_col = None
        self.smiles_col = None
        self.inchi_key_col = None
        self._set_structure_cols()  # Assign mol and SMILES cols using input + defaults
        
        # Current tracked target data
        self._target_col = model['target_col']
        self._target_units = model['input_units']  # Tracks current units of target_col for conversion purposes
        self._average_flag = False
        
        # Convert static -> instance methods on init
        self.chem_cleanup = self._instance_chem_cleanup
        self.handle_duplicates = self._instance_handle_duplicates
        self.final_unit_conversion = self._instance_final_unit_conversion
        self.append_columns = self._instance_append_columns
        self.remove_header_chars = self._instance_remove_header_chars

# ----------------------------------------------------- ACCESSORS ---------------------------------------------------- #

    @property
    def original_df(self) -> pd.DataFrame:
        return self._original_df.copy()

    @property
    def df(self) -> pd.DataFrame:
        return self._df.copy()

    @property
    def model(self) -> dict:
        return copy.deepcopy(self._model)
        
# -------------------------------------------------- PRIVATE METHODS ------------------------------------------------- #
    def _set_structure_cols(self) -> None:
        '''Sets Mol and SMILES columns using declared structure type. Uses defaults if not declared'''
        self.mol_col = self.model['structure_col'] if self.model['structure_type'] == 'mol' else self._default_cols['mol']
        self.smiles_col = self.model['structure_col'] if self.model['structure_type'] == 'smiles' else \
            self._default_cols['smiles']
        self.inchi_key_col = self._default_cols['inchi_key']

    def _drop_na_structures(self) -> None:
        '''Drops NA along declared structure column'''
        self._df.dropna(subset=[self.model['structure_col']], inplace=True)  # NOTE: should save operation to log
        if not len(self._df):
            warnings.warn(bleach_warnings['ALL_NA_STRUCTURES'], RuntimeWarning)

    def _drop_na_targets(self) -> None:
        '''Drops NA along declared target column'''
        run = self.model['format_settings']['remove_na_targets']

        if self._target_col and run and len(self._df):  # If run and TARGET COLUMN DECLARED
            self._df.dropna(subset=[self._target_col], inplace=True)  # NOTE: should save operation to log
            if not len(self._df):
                warnings.warn(bleach_warnings['ALL_NA_TARGETS'], RuntimeWarning)

        elif run:  # If run but not declared target
            warnings.warn(warnings['NA_NO_ACTIVITY_COLUMN'], UserWarning)

    def _build_smiles(self) -> None:
        """Creates a SMILES column in the dataset using dataset MolFile column. DROPS NA."""
        self._df = naclo.dataframes.df_mols_2_smiles(self._df, self.mol_col, self.smiles_col)

    def _build_mols(self) -> None:
        """Creates MolFile column in the dataset using dataset SMILES column. DROPS NA."""
        self._df = naclo.dataframes.df_smiles_2_mols(self._df, self.smiles_col, self.mol_col)
        
    @staticmethod
    def _filter_fragments_factory(filter:str) -> Callable:
        '''Returns a callable SMILES fragment filter function using a key'''
        if filter == 'carbon_count':
            return naclo.fragments.carbon_count
        elif filter == 'mw':
            return naclo.fragments.mw
        elif filter == 'atom_count':
            return naclo.fragments.atom_count
        else:
            raise ValueError('Filter method is not recognized')

    @staticmethod
    def _append_inchi_keys(df:pd.DataFrame, mol_col_name:str, inchi_key_col_name:str) -> pd.DataFrame:
        '''Declares inchi key column name using default. Appends inchi keys to dataset'''
        return naclo.dataframes.df_mols_2_inchi_keys(df, mol_col_name, inchi_key_col_name)

    @staticmethod
    def _drop_columns(df:pd.DataFrame, return_columns:Iterable[str], mol_col_name:str, smiles_col_name:str,
                       inchi_key_col_name:str) -> pd.DataFrame:
        '''Removes columns that the user does not want in the final output.'''
        if 'mol' not in return_columns:
            df = df.drop(mol_col_name, axis=1)
        if 'inchi_key' not in return_columns:
            df = df.drop(inchi_key_col_name, axis=1)
        if 'smiles' not in return_columns:
            df = df.drop(smiles_col_name, axis=1)
        
        return df

    @staticmethod
    def _add_columns(df:pd.DataFrame, columns:Iterable[str], mol_col_name:str) -> pd.DataFrame:
        '''Add columns that the user wants in the final output.'''
        if 'mw' in columns:
            df = df.assign(MW = naclo.mol_weights(df[mol_col_name]))
        return df
    
# -------------------------------------------------- PUBLIC METHODS -------------------------------------------------- #
    
    @staticmethod
    def remove_fragments(df:pd.DataFrame, smiles_col_name:str, mol_col_name:str, desalt:bool,
                          filter_method:Optional[str]) -> pd.DataFrame:
        """Removes salts if specified. Drops NA as a result of salt removal. Filters out other fragments by specified
        method. Regenerates Mols between SMILES operations

        Args:
            df (pd.DataFrame): Dataset
            smiles_col_name (str): Name of SMILES column in df
            mol_col_name (str): Name of Mol column in df
            desalt (bool): Run desalter. If True removes recognized salts
            filter_method (Optional[str]): Method to filter other fragments. Options: Bleach.filter_fragments_methods

        Returns:
            pd.DataFrame: Dataset w/ fragments removed
        """
        df = df.copy()
        f_rebuild_mols = lambda x: naclo.dataframes.df_smiles_2_mols(x, smiles_col_name, mol_col_name)
            
        if desalt:
            df[smiles_col_name] = df[smiles_col_name].apply(naclo.fragments.remove_recognized_salts)  # NOTE: should save operation to log
            df = f_rebuild_mols(df)

            # Drop NA (blank string after salts)
            df = stse.dataframes.convert_to_nan(df, na=[''])  # Convert bc NA is just empty string
            df.dropna(subset=[smiles_col_name], inplace=True)  # Drop NA bc may include molecule that is ONLY salts

        # Filter
        if filter_method and filter_method != 'none':
            df[smiles_col_name] = df[smiles_col_name].apply(Bleach._filter_fragments_factory(filter_method))
            df = f_rebuild_mols(df)
            
        return df
    
    @staticmethod
    def nlm_geometric_mean(df:pd.DataFrame, mol_col_name:str, inchi_key_col_name:str, value_col_name:str,
                           input_units:str, units_col_name:Optional[str]) -> pd.DataFrame:
        """Converts values in df to -log(M) and to computes the geometric mean

        Args:
            df (pd.DataFrame): _description_
            mol_col_name (str): _description_
            inchi_key_col_name (str): _description_
            value_col_name (str): _description_
            input_units (str): _description_
            units_col_name (Optional[str]): _description_

        Returns:
            pd.DataFrame: _description_
        """
        df = df.copy()
        col_name = f'neg_log_molar_{value_col_name}'  # This is necessary to avoid overwriting existing column
        
        # 1. Convert to -log(M)
        if input_units == 'mixed':  # Unit column
            if units_col_name:
                # df[col_name] = UnitConverter(df[value_col_name], df[units_col_name], naclo.mol_weights(df[mol_col_name]))
                df = UnitConverter.df_converter(df=df, mol_col_name=mol_col_name, value_col_name=value_col_name,
                                                units_col_name=units_col_name, output_units='-log(m)', drop_na_units=False)
            else:
                raise ValueError('Must specify units_col_name if input_units is mixed')
        else:
            if units_col_name:
                warnings.warn(bleach_warnings['CONFLICTING_INPUT_UNITS'], UserWarning)
            df[col_name] = UnitConverter.convert(df[value_col_name].astype(float), input_units, '-log(m)')
        
        # 2. Average
        return stse.duplicates.average(df, subsets=[inchi_key_col_name], average_by=col_name)  # NOTE: should save operation to log
    

                
            
            # df[col_name] = UnitConverter.static_neg_log_converter(df[value_col_name], input_units)


# ---------------------------------------------------- MAIN STEPS ---------------------------------------------------- #
    # Step 1
    def drop_na(self) -> None:  # *
        '''Converts blanks to NA. Drops NA Mols or SMILES. Handles NA targets. Removes entire NA columns'''

        # 1. Convert all df blanks and 'none' to NA
        self._df = stse.dataframes.convert_to_nan(self._df)

        # 2. Drop rows
        self._drop_na_structures()
        self._drop_na_targets()

        # 3. Drop entire empty cols
        self._df = stse.dataframes.remove_nan_cols(self._df)  # After dropping rows because columns may BECOME empty
        
    # Step 2 
    def init_structure_compute(self) -> None:  # *
        '''Builds (or rebuilds from Mols) SMILES. Builds Mols if not present in dataset. SMILES should be rebuilt from 
        Mols to canonicalize'''
        if self.model['structure_type'] == 'mol':
            self._build_smiles()
            # Rebuilding Mols not necessary

        elif self.model['structure_type'] == 'smiles':
            self._build_mols()
            self._build_smiles()  # Canonicalize SMILES

    # Step 3
    @staticmethod
    def chem_cleanup(df:pd.DataFrame, smiles_col_name:str, mol_col_name:str, desalt:bool, filter_method:Optional[str],
                    neutralize_charge:bool) -> pd.DataFrame:  # * (except neutralize)
        '''Cleans Mols and SMILES'''
        df = df.copy()

        # 1. Deal with fragments (includes salt step -- may include a molecule that is ONLY salts (NA dropped))
        df = Bleach.remove_fragments(df=df, smiles_col_name=smiles_col_name, mol_col_name=mol_col_name, desalt=desalt,
                                      filter_method=filter_method)

        # 2. Neutralize mols
        if neutralize_charge:
            df[mol_col_name] = naclo.neutralize.neutralize_charges(df[mol_col_name])  # NOTE: should save operation to log
            df = naclo.dataframes.df_mols_2_smiles(df, mol_col_name, smiles_col_name)  # Rebuild SMILES
        
        return df
    
    def _instance_chem_cleanup(self) -> None:
        self._df = Bleach.chem_cleanup(df=self._df, smiles_col_name=self.smiles_col, mol_col_name=self.mol_col,
                                     desalt=self.model['chem_settings']['desalt'],
                                     filter_method=self.model['chem_settings']['fragment_filter'],
                                     neutralize_charge=self.model['chem_settings']['neutralize_charge'])

    # Step 4
    @staticmethod
    def handle_duplicates(df:pd.DataFrame, mol_col_name:str, inchi_key_col_name:str, target_col_name:Optional[str]=None,
                          input_units:Optional[str]=None, units_col_name:Optional[str]=None,
                          method='average') -> pd.DataFrame:  # *
        '''Computes inchi keys. Averages, removes, or keeps duplicates. ONLY BY INCHI KEY FOR NOW. Will return units in
        -log(M) if method is average, else original units'''
        avg_flag = False
        df = Bleach._append_inchi_keys(df, mol_col_name, inchi_key_col_name)

        if method == 'average' and target_col_name:
            # Average only on -log(M)
            df = Bleach.nlm_geometric_mean(df=df, mol_col_name=mol_col_name, inchi_key_col_name=inchi_key_col_name,
                                           value_col_name=target_col_name, input_units=input_units,
                                           units_col_name=units_col_name)  # NOTE: should save operation to log
            avg_flag = True
        elif method == 'remove' or (method == 'average' and not target_col_name):
            if method == 'average':
                warnings.warn(bleach_warnings['DUPLICATES_NO_ACTIVITY_COLUMN'], UserWarning)
            df = stse.duplicates.remove(df, subsets=[inchi_key_col_name])  # NOTE: should save operation to log
        return df, avg_flag
    
    def _instance_handle_duplicates(self) -> None:
        self._df, self._average_flag = Bleach.handle_duplicates(df=self._df, mol_col_name=self.mol_col, inchi_key_col_name=self.inchi_key_col,
                                           target_col_name=self._target_col,
                                           input_units=self.model['input_units'],
                                           units_col_name=self.model['units_col'],
                                           method=self.model['format_settings']['duplicate_compounds'])
        
        # Set units to -log(m) if they were converted for geometric mean
        if self._average_flag:
            self._target_col = f'neg_log_molar_{self._target_col}'  # Change column pointer
            self._target_units = '-log(m)'
    
    @staticmethod  
    def final_unit_conversion(df, input_units, output_units, units_col_name, mol_col_name, value_col_name, drop_na_units):
        if input_units == 'mixed':
            if units_col_name:
                return UnitConverter.df_converter(df=df, mol_col_name=mol_col_name, value_col_name=value_col_name,
                                                units_col_name=units_col_name, output_units=output_units,
                                                drop_na_units=drop_na_units)
            else:
                raise ValueError('Must specify units_col_name if input_units is mixed')
        else:
            if units_col_name:
                warnings.warn(bleach_warnings['CONFLICTING_INPUT_UNITS'], RuntimeWarning)
                
            units_label = 'neg_log_molar' if output_units.lower() == '-log(m)' else output_units  # -log(M) may screw up datasheet characters
            df[f'{units_label}_{value_col_name}'] = UnitConverter.convert(df[value_col_name], input_units, output_units)
            return df
            
    def _instance_final_unit_conversion(self):
        if self._target_units.lower() == self.model['output_units'].lower():
            return  # No change -- already in correct units
        elif not self.model['target_col']:
            warnings.warn(bleach_warnings['OUTPUT_UNITS_CONVERSION'], UserWarning)
            return
        
        # Has already been converted to -log(m)
        if self._average_flag:
            col_name = '{0}_{1}'.format(self.model['output_units'], self.model['target_col'])
            self._df[col_name] = UnitConverter.convert(self._df[self._target_col], '-log(m)', self.model['output_units'])
        else:
            self._df = Bleach.final_unit_conversion(df=self._df, input_units=self.model['input_units'],
                                         output_units=self.model['output_units'],
                                         units_col_name=self.model['units_col'], mol_col_name=self.mol_col,
                                         value_col_name=self._target_col, drop_na_units=self.model['drop_na_units'])

    # Step 6
    @staticmethod
    def append_columns(df:pd.DataFrame, columns:Iterable[str], mol_col_name:str, smiles_col_name:str, inchi_key_col_name:str) -> pd.DataFrame:  # *
        df = df.copy()
        
        df = Bleach._drop_columns(df=df, return_columns=columns, mol_col_name=mol_col_name,
                                  smiles_col_name=smiles_col_name, inchi_key_col_name=inchi_key_col_name)
        df = Bleach._add_columns(df=df, columns=columns, mol_col_name=mol_col_name)
        return df
    
    def _instance_append_columns(self) -> None:
        self._df = Bleach.append_columns(df=self._df, columns=self.model['format_settings']['append_columns'],
                                        mol_col_name=self.mol_col, smiles_col_name=self.smiles_col, inchi_key_col_name=self.inchi_key_col)

    # Step 7
    @staticmethod
    def remove_header_chars(df, chars) -> pd.DataFrame:  # *
        '''Removes any chars listed in a string of chars from the df column headers'''
        return stse.dataframes.remove_header_chars(df, chars)
    
    def _instance_remove_header_chars(self) -> None:
        self._df = Bleach.remove_header_chars(self._df, self.model['format_settings']['remove_header_chars'])


# ----------------------------------------------------- MAIN LOOP ---------------------------------------------------- #
    def main(self) -> pd.DataFrame:
        self.drop_na()  # Before init_structure bc need NA
        self.init_structure_compute()
        self.chem_cleanup()
        self.handle_duplicates()
        
        self.final_unit_conversion()
        
        self.append_columns()
        self.remove_header_chars()
        return self.df



from setuptools import setup, find_packages


setup(
    name='naclo',
    version='0.1.93',
    license='MIT',
    author='Jacob Gerlach',
    author_email='jwgerlach00@gmail.com',
    url='https://github.com/jwgerlach00/naclo',
    description='Cleaning toolset for small molecule drug discovery datasets',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={'naclo': [
        'assets/bleach_model.yml',
        'assets/bleach_warnings.yml',
        'assets/binarize_default_params.json',
        'assets/binarize_default_options.json',
        'assets/recognized_bleach_options.json',
        'assets/recognized_binarize_options.json',
        'assets/recognized_units.json',
        'assets/recognized_salts.json',
    ]},
    python_requires='>=3.6',
    install_requires=[
        'numpy',
        'pandas',
        'PyYAML',
        'rdkit',
        'rdkit_pypi',
        'schema',
        'stse',
    ],
)
